#!/usr/bin/env python
# coding: utf-8

"""\
Code for error parametrization regresison functions:
- CO2 satellite obervations

Equations are from LOGOFLUX-1 report
"""

import numpy as np



def epf(sza, albedo_nir, albedo_swir, aod, cod, cth, error_type='rel', species='CO2'):
    """\

    parameter   | description                       | valid range
    ------------+-----------------------------------+------------------------
    sza         | solar zenith angle (degree)       | 0 - 80 degree
    albedo_nir  | albedo NIR band                   | 0.03 - 0.7
    albedo_swir | albedo SWIR band                  | 0.03 - 0.7
    aod         | aerosol optical depth at 550 nm   | 0 - 0.6
    cod         | cirrus optical depth (NIR)        | 0 - 0.6
    cth         | cirrus top height (km)            | 2 - 20 km
    ------------+-----------------------------------+------------------------
    """

    # either raise if out of range or mask
    #if not (0.0 <= sza <= 80.0):
    #    raise ValueError
    shape = np.shape(sza)


    inc_sza = np.cos(84.0) / np.cos(sza + 9.0) * sza / 75.0
    inc_alb = (1.01 / (albedo_swir + 0.01) - 1.0) * 0.01

    x = np.array([
        np.ones(shape),
        sza - np.full(shape, 50.0),
        albedo_nir - np.full(shape, 0.10),
        albedo_swir - np.full(shape, 0.10),
        aod - np.full(shape, 0.2),
        cod - np.full(shape, 0.05),
        cth - np.full(shape, 10.0),
        aod * inc_sza * inc_alb
    ])

    # CO2/CH4 relative error
    if error_type == 'rel':
        c = {
            'CO2': np.array([1.29121, -0.00220, -0.07275, -1.97335,  -1.00620,  0.34586, -0.00150, 38.86071]),
            'CH4': np.array([9.47532, -0.02678, -7.84754, -11.52154, -7.88635, 15.80352,  0.01702, 275.25335])
        }[species]

    # CO2/CH4 sys. errors
    if error_type == 'sys':
        c = {
            'CO2': np.array([0.30786, 0.01191, 1.97716, -1.55154, -1.40282, 0.70400, -0.03020, -73.46152]),
            'CH4': np.array([1.41086, 0.05545, 9.17830, -6.77504, -9.15046, 5.52796, -0.48784, -335.04555])
        }[species]


    return np.sum(x * c[:,np.newaxis], axis=0)






