#!/usr/bin/env python
# coding: utf-8

from datetime import datetime
import os
import time

import cartopy.crs as ccrs
import netCDF4
import numpy as np
import rasterio
import shapely

import omi

import orbitsim

from . import map
from . import misc
from . import config as cfg


r_earth = 6371.0


DOMAIN = shapely.geometry.Polygon([
    (cfg.startlon, cfg.startlat),
    (cfg.stoplon, cfg.startlat),
    (cfg.stoplon, cfg.stoplat),
    (cfg.startlon, cfg.stoplat)
])

def sunsync_inclination(trev):
    """\
    Calculate inclination of sun-synchronous orbit from revolution time
    (source: https://en.wikipedia.org/wiki/Sun-synchronous_orbit)

    trev :: time for revolution in days
    """
    return np.rad2deg(np.arccos(-(trev * 24. / 3.795)**(7.0/3.0)))


def incl2trev(incl):
    return (-(np.cos(np.deg2rad(incl)) * 3.795**(7.0/3.0)))**(3.0/7.0) / 24.0



def pixel_size(filename, binning=None):
    s = orbitsim.Orbit_sim.from_file(filename)
    s.phase_start = 0.49
    s.phase_end = 0.51
    s.calculate_revolution()
    _, _, lon, lat = reshape_orbit(s)

    dx = [omi.cgrate.geo_distance(lon[10,i,0], lat[10,i,0], lon[10,i,1], lat[10,i,1]) for i in range(s.nrow)]
    dy = [omi.cgrate.geo_distance(lon[10,i,0], lat[10,i,0], lon[10,i,-1], lat[10,i,-1] ) for i in range(s.nrow)]
    dx, dy = np.array(dx), np.array(dy)
    x = np.arange(dx.size) - dx.size / 2.0

    return x, dx, dy


def plot_pixel_size(filename):
    x, dx, dy = pixel_size(filename)

    fig = plt.figure(figsize=(5.0, 3.0))
    ax = fig.add_subplot(111)

    ax.plot(x,dx, '.-', label='across')
    ax.plot(x,dy, '.-', label='along')

    ax.set_xlabel('across-track position')
    ax.set_ylabel('pixel size (km)')
    ax.set_ylim(0, 40)
    ax.grid(True)

    plt.yticks(np.arange(0,41,5))
    plt.legend()
    plt.tight_layout()






def reshape_orbit(s):
    """\
    Reshape orbit lon/lat to along/across position.
    """
    lat = s.lat_center.reshape(-1,s.nrow)
    lon = s.lon_center.reshape(-1,s.nrow)
    lat_corner = s.lat_corner.reshape(-1,s.nrow,4)
    lon_corner = s.lon_corner.reshape(-1,s.nrow,4)
    return lon, lat, lon_corner, lat_corner



def calculate_pixels(scenario_filename, orbit, phase_start=0.33, phase_end=0.37):
    """\
    Calculate ground pixels for satellite scenario.
    """
    s = orbitsim.Orbit_sim.from_file(scenario_filename)
    s.phase_start = orbit + phase_start
    s.phase_end = orbit + phase_end
    s.calculate_revolution()
    return s



def iter_pixel_area(pixel, x0=cfg.startlon, y0=cfg.startlat,
        dx=cfg.dlon, dy=cfg.dlat, ie=cfg.ie, je=cfg.je):

    dx2, dy2 = 0.5 * dx, 0.5 * dy

    # polygon bounds
    minx, miny, maxx, maxy = pixel.bounds

    # round up and down
    minx = np.floor(minx / dx) * dx
    miny = np.floor(miny / dy) * dy
    maxx = np.ceil(maxx / dx) * dx + 0.5 * dx
    maxy = np.ceil(maxy / dy) * dy + 0.5 * dy

    # start point for pixel
    i0 = int(round((minx - x0) / dx))
    j0 = int(round((miny - y0) / dy))

    # iter over grid cells
    for i,x in enumerate(np.arange(minx,maxx,dx), i0):
        for j,y in enumerate(np.arange(miny,maxy,dy), j0):

            if 0 <= i < ie and 0 <= j <= je:
                coords = np.array([
                    [x-dx2, y-dy2],
                    [x-dx2, y+dy2],
                    [x+dx2, y+dy2],
                    [x+dx2, y-dy2]
                ])
                grid_cell = shapely.geometry.Polygon(coords)

                # calculate fraction of intersection and yield
                if grid_cell.intersects(pixel):
                    weight = grid_cell.intersection(pixel).area / grid_cell.area
                    yield j, i, weight


def create_satobs(field, orbit_filename):
    """\
    Create satellite observations from orbit filename.
    """
    _,_,lon,lat = map.calculate_pixel(orbit_filename, orbit_number)
    shape = lon.shape[:2]

    values = np.zeros(shape)
    weights = np.zeros(shape)

    for n,m in np.ndindex(shape):
        print(n,m)

        pixel = create_pixel(lon[n,m], lat[n,m])

        for j,i,weight in iter_pixel_area(pixel):

            values[n,m] += field[j,i] * weight
            weights[n,m] += weight

    return values / weights


def create_pixel(lon, lat, pollon=cfg.pollon, pollat=cfg.pollat):
    """\
    Create pixel polygon in rotpole coords.
    """
    inp = ccrs.PlateCarree()
    out = ccrs.RotatedPole(
        pole_latitude=pollat,
        pole_longitude=pollon
    )
    coords = out.transform_points(inp, lon, lat)
    pixel = shapely.geometry.Polygon(coords)

    return pixel



def save_orbit(orb, scenario_name, orbit_number, output_path):
    """\
    Save orbit to netCDF file.

    name:
     - Sentinel_7_CO2
     - Sentinel_7_NO2_1km
     - Sentinel_7_NO2_2km
     - Sentinel_7_NO2_4km
     - Sentinel_5
    """
    filename = os.path.join(
            output_path,
            '%s_o%03d_l%03d_pixel.nc' % (scenario_name, orbit_number, orb.longitude_eq)
    )

    with netCDF4.Dataset(filename, 'w', format='NETCDF4') as nc:

        # global attributes (add input data)
        nc.setncatts({
            'DESCRIPTION':  'Orbit information for %s instrument' % scenario_name,
            'DATAORIGIN':   ('SRON Orbit Simulator by Joost Aan de Brugh '
                            '(J.M.J.aan.de.Brugh@sron.nl)'),
            'CREATOR':      'Gerrit Kuhlmann',
            'EMAIL':        'gerrit.kuhlmann@empa.ch',
            'AFFILIATION':  'Empa Duebendorf, Switzerland',
            'DATE CREATED': time.ctime(time.time()),
            'STUDY':        'ESA project - Smartcarb',
        })
        nc.createVariable('input', 'c').setncatts({
            # satellite
            'time_for_one_revolution_in_hours': orb.trev * 24.0,
            'satellite_height_in_km': orb.altitude - r_earth,
            'local_solar_time_at_the_northward_equator_pass_in_hours': orb.t_node,
            'orbital_inclination_angle_in_degrees': np.rad2deg(orb.inclination),

            # revolution
            'date': '%04d-%02d-%02d' % tuple(orb.date),
            'longitude_at_daytime_equator_pass_in_degrees': np.rad2deg(orb.longitude_eq),

            # scanlines
            'orbit_phase_at_start_of_scanning': orb.phase_start,
            'orbit_phase_at_end_of_scanning': orb.phase_end,
            'time_per_scan_in_seconds': orb.tsamp,

            # swath
            'number_of_detector_rows': orb.nrow,
            'polynomial_parameters_for_alpha': ', '.join(str(v) for v in orb.polyterms_alpha),
            'polynomial_parameters_for_beta': ', '.join(str(v) for v in orb.polyterms_beta),
        })


        # create dimensions
        nobs = nc.createDimension('nobs', orb.nobs / orb.nrow)
        nrows = nc.createDimension('nrows', orb.nrow)
        ncorner = nc.createDimension('ncorners', orb.ncorner)

        #create variables
        times = nc.createVariable('time', 'f8', ('nobs','nrows'))
        lat = nc.createVariable('latitude', 'f4', ('nobs', 'nrows'))
        lon = nc.createVariable('longitude', 'f4', ('nobs', 'nrows'))
        lat_corner = nc.createVariable('latitude_corners', 'f4', ('nobs', 'nrows', 'ncorners'))
        lon_corner = nc.createVariable('longitude_corners', 'f4', ('nobs', 'nrows', 'ncorners'))
        sza = nc.createVariable('solar_zenith_angle', 'f4', ('nobs', 'nrows'))
        vza = nc.createVariable('viewing_zenith_angle', 'f4', ('nobs', 'nrows'))
        vaz = nc.createVariable('viewing_azimuth_angle', 'f4', ('nobs', 'nrows'))
        saz = nc.createVariable('solar_azimuth_angle', 'f4', ('nobs', 'nrows'))

        # Assign an attribute to the dataset
        sza.standard_name = "solar_zenith_angle"
        sza.long_name = "solar zenith angle"
        sza.units = "degree"
        vza.standard_name = "viewing_zenith_angle"
        vza.long_name = "viewing zenith angle"
        vza.units = "degree"
        vaz.standard_name = "viewing_azimuth_angle"
        vaz.long_name = "viewing azimuth angle"
        vaz.units = "degree"
        saz.standard_name = "solar_azimuth_angle"
        saz.long_name = "solar azimuth angle"
        saz.units = "degree"
        times.standard_name = "time_of_measurement"
        times.long_name = "time of measurement since 2015-01-01 00:00:00 "
        times.units = "seconds since 2015-01-01 00:00:00"
        lat.standard_name = "latitude"
        lat.long_name = "pixel center latitude"
        lat.units = "degrees north"
        lon.standard_name = "longitude"
        lon.long_name = "pixel center longitude"
        lon.units = "degrees east"
        lat_corner.standard_name = "latitude"
        lat_corner.long_name = "pixel corners latitude"
        lat_corner.units = "degrees north"
        lon_corner.standard_name = "longitude"
        lon_corner.long_name = "pixel corners longitude"
        lon_corner.units = "degrees east"

        # Write data
        times[:] = np.array([
                    (datetime(*orb.utc[i]) - datetime(2015,1,1)).total_seconds()
                    for i in range(orb.nobs)
        ]).reshape(-1, orb.nrow)

        lat[:,:] = orb.lat_center.reshape(-1,orb.nrow)
        lon[:,:] = orb.lon_center.reshape(-1,orb.nrow)
        lat_corner[:,:,:] = orb.lat_corner.reshape(-1,orb.nrow,4)
        lon_corner[:,:,:] = orb.lon_corner.reshape(-1,orb.nrow,4)
        sza[:,:] = orb.sza.reshape(-1,orb.nrow)
        vza[:,:] = orb.vza.reshape(-1,orb.nrow)
        saz[:,:] = orb.saz.reshape(-1,orb.nrow)
        vaz[:,:] = orb.vaz.reshape(-1,orb.nrow)




def in_domain(s, domain=DOMAIN):
    """\
    Checks if any pixel center is inside model domain.
    """
    lon, lat, _, _ = reshape_orbit(s)

    #TODO: check for problems with dateline
    lon = np.array([lon[i,j] for i,j in [(0,0), (0,-1), (-1,-1), (-1,0)]])
    lat = np.array([lat[i,j] for i,j in [(0,0), (0,-1), (-1,-1), (-1,0)]])
    orbit = create_pixel(lon, lat)

    return orbit.intersects(domain)







def main(name):
    """\
     - Sentinel_7_CO2
     - Sentinel_7_NO2_1km
     - Sentinel_7_NO2_2km
     - Sentinel_7_NO2_4km
     - Sentinel_5

    """
    scenario_filename = os.path.join(
        '/users/gkuhl/projects/smartcarb/code/config/',
        '%s.txt' % name
    )

    # Sentinel-7
    if name.startswith('Sentinel_7'):
        phase = 0.33, 0.37
        length = 164

    # Sentinel-5
    if name.startswith('Sentinel_5'):
        phase = 0.32, 0.37
        length = 412

    output_path = '/project/d73/output/orbits/%s' % name

    try:
        os.makedirs(output_path)
    except FileExistsError:
        pass

    for orbit_number in range(0,length):
        print(orbit_number)
        s = calculate_pixels(scenario_filename, orbit_number, phase[0], phase[1])
        if in_domain(s):
            save_orbit(s, name, orbit_number, output_path)




if __name__ == '__main__':
    pass







